#!BPY

""" Registration info for Blender menus:
Name: 'LDraw...'
Blender: 234
Group: 'Import'
Tip: 'Import an LDraw (.dat,.ldr,.mpd) file.'
"""

# 2010 : Modification by Yann
# - merging of datimport.py and ldraw.py python script (does not work with 3 scripts)
# - add some "print" to debug : please remove to accelerate script
# - modification of hires and lowres arrays to match new ldraw data tree
# - full fonctionning with ubuntu 10.04 and blender 2.49
# - adding the support of new syntax of Ldraw color Scheme (it required a 
#   modified version of ldconfig.ldr due to limitation of the length of
#   blender materials name (21 characters)
# - modification of the mat.ref value in colorimport.py
#   (0.25 seems to be too low) -> 0.55

# Warning: MPD support and file type detection are not written.
# Currently the first dat is imported with subobjects as joined meshes.
# TODO:
#  Fix matrix stuff. Rotations are somewhat broken for Objects. (Blender internal)
#  Add loading of objects from Blender (children, templates)?
#  Implement material calculation?
#  Complete detection of file type (pieces)
#
#  Flatten works, but fix it to fold in primitives and flatten parts
#  based on which search path they came from(?)

import colorimport
from colorimport import ImporterOptions, ColorImporter

from Blender import NMesh, Object, Text, Material, Scene, Ipo
from Blender.NMesh import Vert, Face
from Blender.Mathutils import Matrix, Vector, MatMultVec, CopyMat, CrossVecs, DotVecs
from Blender.sys import basename, dirsep, dirname, exists, join

import Blender

###########################################################################
# data file management
###########################################################################

#col2bmat={}
# Attempt to load material-color mappings from a Text.
#try:
#  for line in Text.Get('LDrawMaterials').asLines():
#    words=line.split()
#    try:
#      number=int(words[0])
#      name=words[1]
#      col2bmat[number]=Material.Get(name)
#    except IndexError:
#      pass
#except NameError:
#  pass	# Well, no materials defined. What a shame.
#def material(number):
#  '''Return a Blender material for LDraw color'''
#  if col2bmat.has_key(number):
#    return col2bmat[number]
#  # Material 16 (current colour) is useful for parts. The script shouldn't
#  # replace it here as that would break editing of parts.
#  #elif number==16:
#  #  return material(7)
#  else:
#    mat=Material.New('LDraw%d'%(number))
#    try:
#      text=Text.Get('LDrawMaterials')
#    except NameError:
#      text=Text.New('LDrawMaterials')
#    text.write('%d %s\n'%(number,mat.name))
#    col2bmat[number]=mat
#    # TODO: Fill in material properties. (Automix dithered, make transparent..)
#    return mat

class datfile:
  '''datfinder wrapper for a single file'''
  def __init__(self, file):
    self.name=basename(file.name)
    self.file=file
    print "File: ",self.name
    
  def __getitem__(self, name=None):
    if name==self.name or name is None:
      return self.file
    else:
      raise KeyError, name
  def has_key(self, name):
    return name==self.name

class mpdfile:
  '''datfinder wrapper for a multi-part file'''
  def __init__(self, file):
    self.files={}
    first=True
    for line in file:
      if len(line)>7 and line[:7]=='0 FILE ':
	part=[]
	name=line[7:].rstrip()
	if first:
	  self.firstfile=part
	  first=False
	self.files[name]=part
      else:
	part.append(line)
  def __getitem__(self, name=None):
    if name is None:
      return self.firstfile
    else:
      return self.files[name]
  def has_key(self, name):
    return self.files.has_key(name)

class directory:
  '''datfinder wrapper for a directory'''
  def __init__(self, path):
    self.path=path
  def __getitem__(self, name):
    fullname=join(self.path,name)

    if exists(fullname)==1:
      #print "fullName: ",fullname
      return file(fullname)
    else:
      raise KeyError, name
  def has_key(self, name):
    fullname=join(self.path,name)
    print "Testing",fullname
    return exists(fullname)==1

class datfinder:
  '''Class for handling search path for LDraw files. Get line lists as from a dictionary.'''
  def __init__(self):
    self.finders=[]
  def __getitem__(self, name=None):
    for finder in self.finders:
      try:
        return finder[name]
      except KeyError:
        pass
    raise KeyError, name
  def has_key(self, name):
    for finder in self.finders:
      if finder.has_key(name):
        return True
    return False
  def insertpath(self, path):
    '''Add a path (directory, dat, mpd) to the search path. NOTE: This puts it first in the search path.'''
    typ=exists(path)
    if typ==1:
      print "Include: ",path
      # Include the directory for searching
      dir=dirname(path)
      self.finders.insert(0, directory(dir))
      # Check if it's an MPD
      fil=file(path,'rU')
      head=fil.read(7)
      fil.seek(0)
      if head=='0 FILE ':
        # It is
	self.finders.insert(0, mpdfile(fil))
      else:
        # Nope. Assume it's a DAT file.
	self.finders.insert(0, datfile(fil))
    elif typ==2:
      # It's a directory
      self.finders.insert(0, directory(path))
    else:
      raise ValueError, path

def newdatlib(libPath,useHires=False):
  datlib=datfinder()
# parts/b does not exists at home, but models do!
#  hiresArray  = ['p','p/48','parts','parts/s','parts/b']
#  lowresArray = ['p/48','p','parts','parts/s','parts/b']
  hiresArray  = ['p','p/48','parts','parts/s','models']
  lowresArray = ['p/48','p','parts','parts/s','models']
  
  if useHires:
    dataArray = hiresArray
  else:
    dataArray = lowresArray

  for name in dataArray:
    path=join(libPath,name)
    datlib.insertpath(path)
  return datlib

#if __name__ == '__main__':
#  from Blender.Window import FileSelector
#  FileSelector(loaddat)

###########################################################################
# Converter class
###########################################################################

# a little configuration class
class ImporterOptions:
  def __init__(self):
    self.defaults = {'ldrawPath': "/full/path/to/directory/ldraw",'scaleFactor': 0.01, 'useHires':False, 'useSteps':False}
    d=Blender.Registry.GetKey('dat_import',True)
    
    if d == None or d.keys() != self.defaults.keys():
      d = self.defaults
      Blender.Registry.SetKey('dat_import',d,True)
    self.get_ldrawPath = d['ldrawPath']
    self.get_scaleFactor = d['scaleFactor']
    self.get_useHires = d['useHires']
    self.get_useSteps = d['useSteps']

col2bmat={}

SCALE_FACTOR = 0.01
# Attempt to load material-color mappings from a Text.
def loadMaterial():
  for line in Text.Get('LDrawMaterials').asLines():
    print "Blubberlutsh: ",line
    words=line.split()
    try:
      number=int(words[0])
      name=words[1]
      print number,name
      col2bmat[number]=Material.Get(name)
    except IndexError:
      pass


try:
  loadMaterial()
except NameError:
  imp = ColorImporter()
  imp.importColors()
  loadMaterial();
  pass	# Well, no materials defined. What a shame.

def material(number):
  '''Return a Blender material for LDraw color'''
  try:
    return col2bmat[number]
  # Material 16 (current colour) is useful for parts. The script shouldn't
  # replace it here as that would break editing of parts.
  #elif number==16:
  #  return material(7)
  except KeyError:
    mat=Material.New('LDraw%d'%(number))
    try:
      text=Text.Get('LDrawMaterials')
    except NameError:
      text=Text.New('LDrawMaterials')
    text.write('%d %s\n'%(number,mat.name))
    col2bmat[number]=mat
    # TODO: Fill in material properties. (Automix dithered, make transparent..)
    return mat


def getmdl(name):
  '''Gets a legomesh instance for the given model name. May load different files'''
  global models
  # TODO: Determine model type by path. Load paths from environment.
  # Etc etc etc...
  fname=name.replace('\\', dirsep)
  fname=fname.replace('/', dirsep)
  #print "getmdl ",name,"|",fname
  try:
    # Find if we already have a legomesh for that name
    mdl=models[name]
  except KeyError:
    try:
      fil=datlib[name]
    except KeyError:
      fname=fname.lower()
      try:
        fil=datlib[fname]
      except KeyError:
        #fname=fname.upper()
        fil=datlib[fname]
    print "new MeshForName: ",name
    mdl=legomesh(name)
    mdl.fromFile(fil)
    models[name]=mdl
  return mdl

class legomesh:
  '''Class for holding a LDraw model'''

  def __init__(self, name=None):
    self.type=None
    self.name=name
    # If self.mesh is not None, the mesh exists in Blender.
    self.mesh=None
    # Faces contains vertex indicies
    self.faces=[]
    self.facecol=[]
    self.facesmooth=[]
    self.verts=[]
    # Children will be a list of colour, invert, matrix, legomesh
    self.children=[]
    # Winding order None, 'CW' or 'CCW'.
    # Note that we're wound counter-clockwise either way, this affects loading.
    self.windorder=None
    self.invertnext=False

  def toBMesh(self):
    # Note: This means we'll never override a Blender Mesh.
    if self.name and self.mesh is None:
      self.mesh=NMesh.GetRaw(self.name)
    if self.mesh is None:
      mesh=NMesh.New(self.name)
      if self.faces:
	matindex={16: 0}
	bmats=[None]
	for colour in self.facecol:
	  if not matindex.has_key(colour):
	    matindex[colour]=len(bmats)
	    bmats.append(material(colour))
	mesh.materials=bmats
	self.removeDoubles()
	# If this fails, you're probably not using Python 2.3.
	# Change:
	#  Vert(*vert)
	# to:
	#  apply(Vert, vert)
	mesh.verts=[Vert(*vert) for vert in self.verts]
	for face, smooth, colour in map(None, self.faces, self.facesmooth, self.facecol):
	  bface=Face([mesh.verts[v] for v in face])
	  bface.smooth=smooth
	  bface.mat=matindex[colour]
	  mesh.faces.append(bface)
	# Autosmooth threshold will need tuning for some objects.
	# By enabling it only for faces connected to 5 lines, we're pretty safe.
	# Of course, the full job could be done by disconnecting faces at 2 lines
	# by using duplicate vertices, and using ordinary smoothing, I think.
	# Much easier to leave autosmoothing tuning for the user.
	if self.windorder:
	  mesh.setMode('AutoSmooth')
	else:
	  mesh.setMode('TwoSided', 'AutoSmooth')
      self.mesh=mesh
    return self.mesh

  def toBObj(self, mycolour=7):
    '''Creates the Blender objects, including children'''
    global scene
    obj=Object.New('Mesh', self.name)
    obj.link(self.toBMesh())
    obj.colbits=1
    obj.setMaterials([material(mycolour)])
    scene.link(obj)
    self.childrenBObj(obj, mycolour)
    return obj

  def childrenBObj(self, obj, mycolour=7, childlist=None):
    '''Creates the children for our object'''
    if not childlist:
      childlist=self.children
    children=[]
    for colour, invert, matrix, model in childlist:
      if colour==16:
        colour=mycolour
      child=model.toBObj(colour)
      matrix=Matrix(matrix)
      matrix.transpose()
      # FIXME Blender seems to lose some rotations here.
      child.setMatrix(matrix)
      children.append(child)
    obj.makeParent(children, 0, 1)
    return children

  def fromFile(self, file):
#    print "fromFile:",file
    '''Loads LDraw file. Uses iteration.'''
    for line in file:
      self.loadline(line.split())

  def removeDoubles(self):
    if len(self.verts)<=1:
      return
    ind=range(len(self.verts))
    sv=zip(self.verts, ind)
    sv.sort()
    prev_v,first_i = sv.pop(0)
    verts=[prev_v]
    ind[first_i]=prev_newi=0
    for v,i in sv:
      # This is the place to add imprecision.
      # It's also a very cpu intensive loop.
      if v==prev_v:
        # Duplicate vertex
        ind[i]=prev_newi
      else:
        prev_newi+=1
        ind[i]=prev_newi
        prev_v=v
        verts.append(v)
    self.verts=verts
    self.faces=[[ind[v] for v in face] for face in self.faces]

  def flattenParts(self):
    '''Makes sure all Part models used by this model or its submodels are flattened'''
    for colour, invert, matrix, model in self.children:
      if model.type == 'Part' or model.faces:
        model.flatten()
      else:
        model.flattenParts()

  def flatten(self):
    '''Merges all children into the mesh. NOTE: also flattens all children!'''
    if not self.children:
      return
    # Shorthand names to speed this up
    tup=tuple
    mmv=MatMultVec
    vec=Vector
    l=list
    faces=self.faces
    verts=self.verts
    facesmooth=self.facesmooth
    facecol=self.facecol
    for colour, invert, matrix, model in self.children:
      model.flatten()
      if not model.faces:
        continue
      if not model.windorder:
        self.windorder=None
      for col in model.facecol:
        if col==16:
          facecol.append(colour)
        else:
          facecol.append(col)
      voff=len(verts)
      if matrix.determinant()<0:
        invert=not invert
      facesmooth.extend(model.facesmooth)
      if invert:
        for face in model.faces:
          face=[v+voff for v in face]
          face.reverse()
          self.faces.append(face)
      else:
        self.faces.extend([[v+voff for v in f] for f in model.faces])

      self.verts.extend([tup( (matrix*vec(l(v[0:3])+[1])) [:3])for v in model.verts])
    self.removeDoubles()
    self.children=[]

  def step(self):
    '''Records a building step. Does nothing in this class.'''
    pass
  
  def loadline(self, line):
#    print "loadLine: ",line
    options = ImporterOptions()
    '''Loads a single line (passed as list of words) from an LDraw file'''
    if line:
      if line[0] == '0':
        try:
          if line[1] == 'STEP':
            self.step()
          elif line[1] == 'BFC':
	           # Back face culling extension - we use this for winding order
            if not (self.faces or self.children):
              # Okay, this BFC command MAY be allowed to set winding order.
              if 'CW' in line[2:]:
                self.windorder='CW'
              else:
                self.windorder='CCW'
            if 'INVERTNEXT' in line[2:]:
              self.invertnext=True
          elif line[1] in ('LDRAW_ORG', 'Unofficial', 'Un-official'):
            self.type=line[2]
          elif tuple(line[1:3]) in (('Official', 'LCAD'), ('Official', 'LCad'), ('Original', 'LDraw')):
            self.type=line[3]
        except IndexError:
          pass
      elif line[0] == '1':
        # Inclusion of another file
        x,y,z,a,b,c,d,e,f,g,h,i=map(float, line[2:14])
        #mat=map(float,line[2:14])
        #mat=Matrix([a,b,c,x],[d,e,f,y],[g,h,i,z],[0,0,0,1])
        mat=Matrix([a,b,c,x*options.get_scaleFactor],[d,e,f,y*options.get_scaleFactor],[g,h,i,z*options.get_scaleFactor],[0,0,0,1])

        # Lousy docs. MatMultVec only works if the Matrix is square;
        # otherwise, it will break, including crashes.
        #mat=Matrix(mat[3:6]+[mat[0]], mat[6:9]+[mat[1]], mat[9:]+[mat[2]],[0,0,0,1])
        name=line[14]
        colour=int(line[1])
        childmdl=getmdl(name)
        if childmdl.type in ('Primitive', 'Subpart'):
          self.type='Part'
        self.children.append((colour, self.invertnext, mat, childmdl))
        self.invertnext=False
      elif line[0] in ('3','4'): 
        verts=[ map(float, line[i:i+3]) for i in range(2, 2+int(line[0])*3, 3)]

        # scale the vertices
        tverts = []
        for vertex in verts:
          vec = Vector(vertex)
          vec = vec * options.get_scaleFactor
          tverts.append(vec);
        verts = tverts
        if self.windorder=='CW':
          verts.reverse()
        elif self.windorder is None and len(verts)==4:
          # Fix bowtie quads
          v=[Vector(v) for v in verts]
          # This should be a diagonal of the quad
          v2m0=v[2]-v[0]
          # The normal of the plane
          n=CrossVecs(v[1]-v[0], v2m0)
          # If we want to verify that the quad is flat:
          #k=DotVecs(v[0],n)
          #if DotVecs(v[3],n) != k:
          #  raise "Not Coplanar"
          # Get a vector pointing along the plane perpendicular to the diagonal
          n2=CrossVecs(n, v2m0)
          # Calculate the position relative to the diagonal of three corners
          k=[DotVecs(p,n2) for p in v[1:]]
          # k[1] is *on* the diagonal, relative 0.
          # Check if the corners are on opposite sides of the diagonal
          if cmp(k[1],k[0]) == cmp(k[1],k[2]):
            #print "Bowtie quad",v
            verts=[verts[i] for i in 0,2,3,1]
        first=len(self.verts)
        face=range(first, first+len(verts))
        self.verts.extend(map(tuple, verts))
        self.faces.append(face)
        self.facecol.append(int(line[1]))
        self.facesmooth.append(0)
      elif line[0] == '5':
        # Conditional line, used on curved surfaces.
        v=[tuple(map(float, line[2:5])), tuple(map(float, line[5:8]))]
        try:
          v=[self.verts.index(x) for x in v]
        except ValueError:
          # This line isn't among our faces
          return
        # Find faces with that edge and set them smooth.
        for i in xrange(len(self.faces)):
          face=self.faces[i]
          if v[0] in face and v[1] in face:
            self.facesmooth[i]=1

class maindat(legomesh):
  '''LDR loader - implements STEP support, but not for submodels.'''

  def __init__(self, name=None):
    legomesh.__init__(self, name)
    self.curframe=1
    self.steps=[]

  def fromFile(self, file):
#    print "fromFile:",file
    legomesh.fromFile(self, file)
    if self.children:
      self.step()
      self.curframe-=1

  def step(self):
    '''Records the children as belonging to one step'''
    if not self.children: return
    self.steps.append(self.children)
    self.children=[]

  def flattenParts(self):
    '''Makes sure all Part models used by this model or its submodels are flattened'''
    for children in self.steps:
      for colour, invert, matrix, model in children:
        # I don't know why, but some Parts just lack the Part word.
	if model.type in ('Part', 'update'):
	  model.flatten()
	else:
	  model.flattenParts()

  def childrenBObj(self, obj, mycolour=7):
    '''Creates the children for our object, including step animations'''
    allchildren=[]
    for step in xrange(1, len(self.steps)+1):
      # Blender 2.34 can't edit an Object Ipo until it has been assigned to an Object.
      children=legomesh.childrenBObj(self, obj, mycolour, self.steps[step-1])
      allchildren.extend(children)
      if ImporterOptions().get_useSteps == True:
        ipo=Ipo.New('Object', 'step%d'%step)
        for child in children:
          child.setIpo(ipo)
        ipo.addCurve('Layer')
        c=ipo.getCurves()[0]
        if step!=1:
          c.addBezier((1, 1))
        c.addBezier((step, 3))
    return allchildren

def loaddat(filename):
  global scene, models, datlib
  scene=Scene.getCurrent()
  options = ImporterOptions();
  datlib=newdatlib(options.get_ldrawPath,options.get_useHires)
  datlib.insertpath(filename)
  models={}
  name=basename(filename)
  lm=maindat(name)
  # None will cause it to load the first part of an MPD.
  lm.fromFile(datlib[None])
  lm.flattenParts()
  steps=len(lm.steps)
  if steps:
    rc=scene.getRenderingContext()
    rc.currentFrame(steps)
    rc.endFrame(steps)
  obj=lm.toBObj()
  obj.RotX=3.1415*-90/180
  scene.update()
  # For some reason, the scene update doesn't do this.
  Blender.Redraw()

if __name__ == '__main__':
  from Blender.Window import FileSelector
  FileSelector(loaddat)

